# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.io.remote`."""

from pathlib import PureWindowsPath

import pytest

from ...testing.errors import pytest_skip_flaky_network
from .. import remote as io_remote

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

REMOTE_FILE = "https://gitlab.com/gwpy/gwpy/-/raw/main/README.md"
README_HEAD = "GWpy is a collaboration-driven Python package"


@pytest.mark.parametrize(("url", "result"), [
    pytest.param("/path/to/data", False, id="path"),
    pytest.param("file:///path/to/data", False, id="file"),
    pytest.param("https://example.com/path/to/data", True, id="http"),
    pytest.param("pelican://example.com/path/to/data", True, id="pelican"),
    pytest.param(PureWindowsPath(r"c:\Users\myname\mydata.txt"), False, id="windows"),
])
def test_is_remote(url, result):
    """Test `is_remote()`."""
    assert io_remote.is_remote(url) is result


@pytest_skip_flaky_network
def test_download_file():
    """Test `download_file()`."""
    path = io_remote.download_file(REMOTE_FILE, cache=False)
    with open(path) as file:
        assert next(file).strip().startswith(README_HEAD)


@pytest_skip_flaky_network
def test_open_remote_file():
    """Test `open_remote_file()`."""
    with io_remote.open_remote_file(
        REMOTE_FILE,
        cache=False,
        encoding="utf-8",
    ) as file:
        assert next(file).strip().startswith(README_HEAD)


@pytest_skip_flaky_network
@pytest.mark.requires("requests_pelican")
def test_download_file_pelican():
    """Test `download_file()` with a ``pelican://`` URI."""
    path = io_remote.download_file(
        "osdf:///gwdata/zenodo/README.zenodo",
        cache=False,
    )
    with open(path) as file:
        assert next(file).strip() == "## Mirror of IGWN Zenodo Communities"


@pytest_skip_flaky_network
@pytest.mark.requires("requests_pelican")
def test_open_remote_file_pelican():
    """Test `open_remote_file()` with a ``pelican://`` URI."""
    with io_remote.open_remote_file(
        "osdf:///gwdata/zenodo/README.zenodo",
        cache=False,
        encoding="utf-8",
    ) as file:
        assert next(file).strip() == "## Mirror of IGWN Zenodo Communities"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extension of the basic matplotlib Figure for GWpy."""

from __future__ import annotations

import contextlib
import importlib
import itertools
from collections.abc import (
    KeysView,
    ValuesView,
)
from itertools import zip_longest
from typing import TYPE_CHECKING

import numpy
from matplotlib import (
    _pylab_helpers,
    backends,
    figure,
    get_backend,
)
from matplotlib.artist import setp
from matplotlib.gridspec import GridSpec
from matplotlib.projections import get_projection_class

from . import (
    colorbar as gcbar,
    utils,
)
from .gps import GPS_SCALES
from .rc import (
    MPL_RCPARAMS,
    get_subplot_params,
    rcParams,
)

if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType
    from typing import (
        BinaryIO,
        Literal,
    )

    from astropy.units import UnitBase
    from matplotlib.axes import Axes
    from matplotlib.collections import Collection
    from matplotlib.colorbar import Colorbar
    from matplotlib.image import AxesImage
    from numpy.typing import ArrayLike

    from ..segments import DataQualityFlag
    from .segments import SegmentAxes

__all__ = ["Plot"]

try:
    IPYTHON = __IPYTHON__
except NameError:
    IPYTHON = False

IterableType = list | tuple | dict | KeysView | ValuesView


def interactive_backend() -> bool:
    """Return `True` if the current backend is interactive."""
    try:
        interactive_bk = backends.backend_registry.list_builtin(
            backends.BackendFilter.INTERACTIVE,
        )
    except AttributeError:  # matplotlib < 3.9.0
        from matplotlib.rcsetup import interactive_bk
    return get_backend() in interactive_bk


def get_backend_mod(name: str | None = None) -> ModuleType:
    """Return the imported module for the given backend name.

    Parameters
    ----------
    name : `str`, optional
        the name of the backend, defaults to the current backend.

    Returns
    -------
    backend_mod: `module`
        the module as returned by :func:`importlib.import_module`

    Examples
    --------
    >>> from gwpy.plot.plot import get_backend_mod
    >>> print(get_backend_mod('agg'))
    <module 'matplotlib.backends.backend_agg' from ... >
    """
    if name is None:
        name = get_backend()
    backend_name = (
        name[9:]
        if name.startswith("module://")
        else f"matplotlib.backends.backend_{name.lower()}"
    )
    return importlib.import_module(backend_name)


class Plot(figure.Figure):
    """An extension of the core matplotlib `~matplotlib.figure.Figure`.

    The `Plot` provides a number of methods to simplify generating
    figures from GWpy data objects, and modifying them on-the-fly in
    interactive mode.
    """

    def __init__(
        self,
        *data: ArrayLike | IterableType,
        **kwargs,
    ) -> None:
        """Initialise a new `Plot`."""
        # get default x-axis scale if all axes have the same x-axis units
        kwargs.setdefault(
            "xscale",
            _parse_xscale(_group_axes_data(data, flat=True)),
        )

        # set default size for time-axis figures
        if (
            kwargs.get("projection") == "segments"
            or kwargs.get("xscale") in GPS_SCALES
        ):
            kwargs.setdefault("figsize", (12, 6))
            kwargs.setdefault("xscale", "auto-gps")

        # initialise figure
        figure_kw = {
            key: kwargs.pop(key)
            for key in utils.FIGURE_PARAMS
            if key in kwargs
        }
        self._init_figure(**figure_kw)

        # initialise axes with data
        if data or kwargs.get("geometry"):
            self._init_axes(data, **kwargs)

    def _init_figure(self, **kwargs) -> None:
        from matplotlib import pyplot

        # create Figure
        num = kwargs.pop("num", max(pyplot.get_fignums() or {0}) + 1)
        self._parse_subplotpars(kwargs)
        super().__init__(**kwargs)

        # add interactivity (scraped from pyplot.figure())
        backend_mod = get_backend_mod()
        try:
            manager = backend_mod.new_figure_manager_given_figure(num, self)
        except AttributeError:
            upstream_mod = importlib.import_module(
                pyplot.new_figure_manager.__module__,
            )
            canvas = upstream_mod.FigureCanvasBase(self)
            manager = upstream_mod.FigureManagerBase(canvas, 1)
        manager._cidgcf = manager.canvas.mpl_connect(
            "button_press_event",
            lambda ev: _pylab_helpers.Gcf.set_active(manager),
        )
        _pylab_helpers.Gcf.set_active(manager)
        pyplot.draw_if_interactive()

    def _init_axes(
        self,
        data: ArrayLike | IterableType,
        method: str = "plot",
        *,
        xscale: str | None = None,
        sharex: str | bool = False,
        sharey: str | bool = False,
        geometry: tuple[int, int] | None = None,
        separate: bool | None = None,
        **kwargs,
    ) -> list[Axes]:
        """Populate this figure with data, creating `Axes` as necessary."""
        if isinstance(sharex, bool):
            sharex = "all" if sharex else "none"
        if isinstance(sharey, bool):
            sharey = "all" if sharey else "none"

        # parse keywords
        axes_kw = {
            key: kwargs.pop(key)
            for key in utils.AXES_PARAMS
            if key in kwargs
        }

        # handle geometry and group axes
        if geometry is not None and geometry[0] * geometry[1] == len(data):
            separate = True
        axes_groups = _group_axes_data(data, separate=separate)
        if geometry is None:
            geometry = (len(axes_groups), 1)
        nrows, ncols = geometry
        if axes_groups and nrows * ncols != len(axes_groups):
            # mismatching data and geometry
            msg = (
                f"cannot group data into {len(axes_groups)} with "
                f"a {nrows}x{ncols} grid"
            )
            raise ValueError(msg)

        # create grid spec
        gs = GridSpec(nrows, ncols)
        axarr = numpy.empty((nrows, ncols), dtype=object)

        # set default labels
        defxlabel = "xlabel" not in axes_kw
        defylabel = "ylabel" not in axes_kw
        flatdata = [s for group in axes_groups for s in group]
        for ax in ("x", "y"):
            unit = _common_axis_unit(flatdata, axis=ax)
            if unit:
                axes_kw.setdefault(
                    f"{ax}label",
                    unit.to_string("latex_inline_dimensional"),
                )

        # create axes for each group and draw each data object
        for group, (row, col) in zip_longest(
            axes_groups,
            itertools.product(range(nrows), range(ncols)),
            fillvalue=[],
        ):
            # create Axes
            shared_with = {
                "none": None,
                "all": axarr[0, 0],
                "row": axarr[row, 0],
                "col": axarr[0, col],
            }
            axes_kw["sharex"] = shared_with[sharex]
            axes_kw["sharey"] = shared_with[sharey]
            axes_kw["xscale"] = xscale if xscale else _parse_xscale(group)
            ax = axarr[row, col] = self.add_subplot(gs[row, col], **axes_kw)

            # plot data
            plot_func = getattr(ax, method)
            if method in ("imshow", "pcolormesh"):
                for obj in group:
                    plot_func(obj, **kwargs)
            elif group:
                plot_func(*group, **kwargs)

            # set default axis labels
            for axis, share, pos, n, def_ in (
                (ax.xaxis, sharex, row, nrows, defxlabel),
                (ax.yaxis, sharey, col, ncols, defylabel),
            ):
                # hide label if shared axis and not bottom left panel
                if share == "all" and pos < n - 1:
                    axis.set_label_text("")
                # otherwise set default status
                else:
                    axis.isDefault_label = def_

        return self.axes

    @staticmethod
    def _parse_subplotpars(kwargs: dict) -> None:
        """Parse and set ``"subplotpars"`` from ``kwargs``."""
        # dynamically set the subplot positions based on the figure size
        # -- only if the user hasn't customised the subplot params
        figsize = kwargs.get("figsize") or rcParams["figure.figsize"]
        subplotpars = get_subplot_params(figsize)
        use_subplotpars = (
            "subplotpars" not in kwargs
            and all(
                rcParams[f"figure.subplot.{pos}"]
                == MPL_RCPARAMS[f"figure.subplot.{pos}"]
                for pos in ("left", "bottom", "right", "top")
            )
        )
        if use_subplotpars:
            kwargs["subplotpars"] = subplotpars

    # -- Plot methods ----------------

    def refresh(self) -> None:
        """Refresh the current figure."""
        self.canvas.draw()

    def show(
        self,
        warn: bool = True,
        block: bool | None = None,
    ) -> None:
        """Display the current figure (if possible).

        If blocking, this method replicates the behaviour of
        :func:`matplotlib.pyplot.show()`, otherwise it just calls up to
        :meth:`~matplotlib.figure.Figure.show`.

        This method also supports repeatedly showing the same figure, even
        after closing the display window, which isn't supported by
        `pyplot.show` (AFAIK).

        Parameters
        ----------
        block : `bool`, optional
            open the figure and block until the figure is closed, otherwise
            open the figure as a detached window, default: `None`.
            If `None`, block if using an interactive backend and _not_
            inside IPython.

        warn : `bool`, optional
            print a warning if matplotlib is not running in an interactive
            backend and cannot display the figure, default: `True`.
        """
        # this method tries to reproduce the functionality of pyplot.show,
        # mainly for user convenience. However, as of matplotlib-3.0.0,
        # pyplot.show() ends up calling _back_ to Plot.show(),
        # so we have to be careful not to end up in a recursive loop
        import inspect
        try:
            callframe = inspect.currentframe().f_back
        except AttributeError:
            pass
        else:
            if "matplotlib" in callframe.f_code.co_filename:
                block = False

        # render
        super().show(warn=warn)

        # don't block on ipython with interactive backends
        if block is None and interactive_backend():
            block = not IPYTHON

        # block in GUI loop (stolen from mpl.backend_bases._Backend.show)
        if block:
            backend_mod = get_backend_mod()
            backend_mod.Show().mainloop()

    def save(self, fname: str | Path | BinaryIO, **kwargs) -> None:
        """Save the figure to disk.

        This method is an alias to :meth:`~matplotlib.figure.Figure.savefig`,
        all arguments are passed directory to that method.
        """
        self.savefig(fname, **kwargs)

    def close(self) -> None:
        """Close the plot and release its memory."""
        from matplotlib.pyplot import close
        for ax in self.axes[::-1]:
            # avoid matplotlib/matplotlib#9970
            ax.set_xscale("linear")
            ax.set_yscale("linear")
            # clear the axes
            ax.cla()
        # close the figure
        close(self)

    # -- axes manipulation -----------

    def get_axes(self, projection: str | None = None) -> list[Axes]:
        """Find all `Axes`, optionally matching the given projection.

        Parameters
        ----------
        projection : `str`
            Name of axes types to return.

        Returns
        -------
        axlist : `list` of `~matplotlib.axes.Axes`
            The `Axes` that match the given ``projection``.
        """
        if projection is None:
            return self.axes
        return [ax for ax in self.axes if ax.name == projection.lower()]

    # -- colour bars ----------------------------

    def colorbar(
        self,
        mappable: Collection | AxesImage | None = None,
        *,
        cax: Axes | None = None,
        ax: Axes | None = None,
        fraction: float = 0.,
        emit: bool = True,
        **kwargs,
    ) -> Colorbar:
        """Add a colorbar to the current `Plot`.

        This method differs from the default
        :meth:`matplotlib.figure.Figure.colorbar` in that it doesn't
        resize the parent `Axes` to accommodate the colorbar, but rather
        draws a new Axes alongside it.

        Parameters
        ----------
        mappable : `~matplotlib.image.AxesImage`, `~matplotlib.collections.Collection`, optional
            Collection against which to map the colorbar.
            Default is the most-recently-added mappable artist.

        cax : `~matplotlib.axes.Axes`, optional
            Axes on which to draw colorbar.
            By default a new `Axes` will be created.

        ax : `~matplotlib.axes.Axes`
            Axes relative to which to position colorbar.
            The default is the `Axes` containing the ``mappable``.

        fraction : `float`, optional
            Fraction of original axes to use for colorbar.
            The default (``fraction=0``) is to not resize the
            original axes at all.

        emit : `bool`, optional
            If `True` update all mappables on `Axes` to match the same
            colouring as the colorbar.

        kwargs
            Other keyword arguments are passed to
            :meth:`~matplotlib.figure.Figure.colorbar`.

        Returns
        -------
        cbar : `~matplotlib.colorbar.Colorbar`
            The newly added `Colorbar`.

        Notes
        -----
        To revert to the default matplotlib behaviour, pass ``fraction=0.15``.

        See Also
        --------
        matplotlib.figure.Figure.colorbar
        matplotlib.colorbar.Colorbar

        Examples
        --------
        >>> import numpy
        >>> from gwpy.plot import Plot

        To plot a simple image and add a colorbar:

        >>> plot = Plot()
        >>> ax = plot.gca()
        >>> ax.imshow(numpy.random.randn(120).reshape((10, 12)))
        >>> plot.colorbar(label='Value')
        >>> plot.show()

        Colorbars can also be generated by directly referencing the parent
        axes:

        >>> Plot = Plot()
        >>> ax = plot.gca()
        >>> ax.imshow(numpy.random.randn(120).reshape((10, 12)))
        >>> ax.colorbar(label='Value')
        >>> plot.show()
        """
        return gcbar.colorbar(
            self,
            ax,
            mappable,
            cax=cax,
            fraction=fraction,
            emit=emit,
            **kwargs,
        )

    # -- extra methods ---------------

    def add_segments_bar(
        self,
        segments: DataQualityFlag,
        ax: Axes | None = None,
        height: float = 0.14,
        pad: float = 0.1,
        *,
        sharex: bool = True,
        location: Literal["top", "bottom"] = "bottom",
        **plotargs,
    ) -> SegmentAxes:
        """Add a segment bar `Plot` indicating state information.

        By default, segments are displayed in a thin horizontal set of Axes
        sitting immediately below the x-axis of the main,
        similarly to a colorbar.

        Parameters
        ----------
        segments : `~gwpy.segments.DataQualityFlag`
            A data-quality flag, or `SegmentList` denoting state segments
            about this Plot.

        ax : `Axes`, optional
            Specific `Axes` relative to which to position new `Axes`,
            defaults to :func:`~matplotlib.pyplot.gca()`.

        height : `float`, optional
            Height of the new axes, as a fraction of the anchor axes.

        pad : `float`, optional
            Padding between the new axes and the anchor, as a fraction of
            the anchor axes dimension.

        sharex : `True`, `~matplotlib.axes.Axes`, optional
            Either `True` to set ``sharex=ax`` for the new segment axes,
            or an `Axes` to use directly.

        location : `str`, optional
            Location for new segment axes, defaults to ``'bottom'``,
            acceptable values are ``'top'`` or ``'bottom'``.

        plotargs
            Other keyword arguments are passed to
            :meth:`~gwpy.plot.SegmentAxes.plot`.

        See Also
        --------
        SegmentAxes
            For more details about segment axes.
        """
        # get axes to anchor against
        if not ax:
            ax = self.gca()

        # set options for new axes
        axes_kw = {
            "pad": pad,
            "sharex": ax if sharex is True else sharex or None,
            "axes_class": get_projection_class("segments"),
        }

        # map X-axis limit from old axes
        if axes_kw["sharex"] is ax and not ax.get_autoscalex_on():
            axes_kw["xlim"] = ax.get_xlim()

        # if axes uses GPS scaling, copy the epoch as well
        with contextlib.suppress(AttributeError):
            axes_kw["epoch"] = ax.get_epoch()

        # add new axes
        try:
            divider = ax.get_axes_locator()._axes_divider
        except AttributeError:
            # get_axes_locator() is None _or_ the _axes_divider property
            # has been removed
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            divider = make_axes_locatable(ax)

        if location not in {"top", "bottom"}:
            msg = "segments can only be positoned at 'top' or 'bottom'"
            raise ValueError(msg)

        # create axes
        segax = divider.append_axes(location, height, **axes_kw)

        # update anchor axes
        if axes_kw["sharex"] is ax and location == "bottom":
            # map label
            segax.set_xlabel(ax.get_xlabel())
            segax.xaxis.isDefault_label = ax.xaxis.isDefault_label
            ax.set_xlabel("")
            # hide ticks on original axes
            setp(ax.get_xticklabels(), visible=False)

        # plot segments
        segax.plot(segments, **plotargs)
        segax.grid(
            visible=False,
            which="both",
            axis="y",
        )
        segax.autoscale(axis="y", tight=True)

        return segax


# -- utilities -----------------------

def _group_axes_data(
    inputs: IterableType,
    *,
    separate: bool | None = None,
    flat: bool = False,
) -> list[list[ArrayLike]]:
    """Determine the number of axes from the input args to this `Plot`.

    Parameters
    ----------
    inputs : `list` of array-like data sets
        A list of data arrays, or a list of lists of data sets.

    separate : `bool`, optional
        Plot each set of data on a separate `Axes`.

    flat : `bool`, optional
        Return a flattened list of data objects.

    Returns
    -------
    axesdata : `list` of lists of array-like data
        A `list` with one element per required `Axes` containing the
        array-like data sets for those `Axes`, unless ``flat=True``
        is given.

    Notes
    -----
    The logic for this method is as follows:

    - if a `list` of data arrays are given, and `separate=False`, use 1 `Axes`
    - if a `list` of data arrays are given, and `separate=True`, use N `Axes,
      one for each data array
    - if a nested `list` of data arrays are given, ignore `sep` and
      use one `Axes` for each group of arrays.

    Examples
    --------
    >>> from gwpy.plot import Plot
    >>> Plot._group_axes_data([1, 2], separate=False)
    [[1, 2]]
    >>> Plot._group_axes_data([1, 2], separate=True)
    [[1], [2]]
    >>> Plot._group_axes_data([[1, 2], 3])
    [[1, 2], [3]]
    """
    # determine auto-separation
    if separate is None and inputs:
        separate = bool(
            # if given a nested list of data, multiple axes are required
            any(isinstance(x, IterableType) for x in inputs)
            # if data are of different types, default to separate
            or not all(type(x) is type(inputs[0]) for x in inputs),
        )

    # build list of lists
    out = []
    for x in inputs:
        if isinstance(x, dict):  # unwrap dict
            x = list(x.values())  # noqa: PLW2901

        # new group from iterable, notes:
        #     the iterable is presumed to be a list of independent data
        #     structures, unless its a list of scalars in which case we
        #     should plot them all as one
        if (
            isinstance(x, KeysView | ValuesView)
            or (
                isinstance(x, list | tuple)
                and (
                    not x
                    or not numpy.isscalar(x[0])
                )
            )
        ):
            out.append(x)

        # dataset starts a new group
        elif separate or not out:
            out.append([x])

        # dataset joins current group
        else:  # append input to most recent group
            out[-1].append(x)

    if flat:
        return [s for group in out for s in group]

    return out


def _common_axis_unit(data: IterableType, axis: str = "x") -> UnitBase | None:
    """Parse a common unit for the given ``axis`` for these ``data``."""
    units = set()
    uname = f"{axis}unit"
    for x in data:
        units.add(getattr(x, uname, None))
    if len(units) == 1:
        return units.pop()
    return None


def _parse_xscale(data: IterableType) -> str | None:
    """Parse the x-axis scale that should be applied to these ``data``."""
    unit = _common_axis_unit(data, axis="x")
    if unit is None:
        return None
    if unit.physical_type == "time":
        return "auto-gps"
    return None



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""A spectral-variation histogram class."""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

import numpy
from astropy.units import Quantity

from ..io.registry import UnifiedReadWriteMethod
from ..segments import Segment
from ..types import Array2D
from ..types.sliceutils import null_slice
from ..utils.misc import property_alias
from . import FrequencySeries
from .connect import (
    SpectralVarianceRead,
    SpectralVarianceWrite,
)

if TYPE_CHECKING:
    from typing import (
        ClassVar,
        Never,
        Self,
    )

    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike
    from numpy.typing import (
        ArrayLike,
        NDArray,
    )

    from ..detector import Channel
    from ..plot import Plot
    from ..spectrogram import Spectrogram
    from ..time import SupportsToGps
    from ..types import Series
    from ..typing import (
        Array1D,
        ArrayLike1D,
        UnitLike,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = ["SpectralVariance"]


def _default_bins(
    data: NDArray,
    nbins: int,
    low: float | None,
    high: float | None,
    unit: UnitBase,
    *,
    log: bool,
) -> Array1D:
    if low is None:
        low = data.min() / 2
    if high is None:
        high = data.max() * 2

    if isinstance(low, Quantity):
        low = low.to(unit).value
    if isinstance(high, Quantity):
        high = high.to(unit).value

    if log:
        low = numpy.log10(low)
        high = numpy.log10(high)

    if log:
        return numpy.logspace(low, high, num=nbins + 1)
    return numpy.linspace(low, high, num=nbins + 1)


class SpectralVariance(Array2D):
    """A variance histogram of a `FrequencySeries`."""

    _metadata_slots: ClassVar[tuple[str, ...]] = (
        *FrequencySeries._metadata_slots,  # noqa: SLF001
        "bins",
    )
    _default_xunit: ClassVar[UnitBase] = FrequencySeries._default_xunit  # noqa: SLF001
    _rowclass: ClassVar[type[Series]] = FrequencySeries

    def __new__(
        cls,
        data: ArrayLike,
        bins: ArrayLike1D,
        unit: UnitLike = None,
        f0: Quantity | float | None = None,
        df: Quantity | float | None = None,
        frequencies: ArrayLike1D | None = None,
        name: str | None = None,
        channel: Channel | str | None = None,
        epoch: SupportsToGps | None = None,
        **kwargs,
    ) -> Self:
        """Generate a new SpectralVariance histogram."""
        # parse x-axis params
        if f0 is not None:
            kwargs["x0"] = f0
        if df is not None:
            kwargs["dx"] = df
        if frequencies is not None:
            kwargs["xindex"] = frequencies

        # generate SpectralVariance using the Series constructor
        new = super(Array2D, cls).__new__(
            cls,
            data,
            unit=unit,
            name=name,
            channel=channel,
            epoch=epoch,
            **kwargs,
        )

        # set bins
        new.bins = bins

        return new

    # -- properties ------------------

    @property
    def bins(self) -> Quantity:
        """Array of bin edges, including the rightmost edge.

        :type: `astropy.units.Quantity`
        """
        return self._bins

    @bins.setter
    def bins(self, bins: QuantityLike | None) -> None:
        if bins is None:
            del self.bins
            return
        bins = Quantity(bins)
        if bins.size != self.shape[1] + 1:
            msg = (
                "SpectralVariance.bins must be given as a list of bin edges, "
                "including the rightmost edge, and have length 1 greater than "
                "the y-axis of the SpectralVariance data"
            )
            raise ValueError(msg)
        self._bins = bins

    @bins.deleter
    def bins(self) -> None:
        with suppress(AttributeError):
            del self._bins

    # over-write yindex and yspan to communicate with bins
    @property  # type: ignore[misc]
    def yindex(self) -> Quantity:
        """List of left-hand amplitude bin edges."""
        return self.bins[:-1]

    @property
    def yspan(self) -> Segment:
        """Amplitude range (low, high) spanned by this array."""
        return Segment(self.bins.value[0], self.bins.value[-1])

    @property  # type: ignore[misc]
    def dy(self) -> Quantity:
        """Size of the first (lowest value) amplitude bin."""
        return self.bins[1] - self.bins[0]

    @property  # type: ignore[misc]
    def y0(self) -> Quantity:
        """Starting value of the first (lowest value) amplitude bin."""
        return self.bins[0]

    f0 = property_alias(
        Array2D.x0,  # type: ignore[arg-type]
        doc="Starting frequency for this `SpectralVariance`.",
    )
    df = property_alias(
        Array2D.dx,  # type: ignore[arg-type]
        doc="Frequency spacing of this `SpectralVariance`.",
    )
    frequencies = property_alias(
        Array2D.xindex,  # type: ignore[arg-type]
        doc="Array of frequencies for each sample",
    )

    @property
    def T(self) -> Never:  # noqa: N802
        """Transpose is not supported."""
        msg = f"transposing a {type(self).__name__} is not supported"
        raise NotImplementedError(msg)

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(SpectralVarianceRead)
    write = UnifiedReadWriteMethod(SpectralVarianceWrite)

    # -- methods ---------------------

    def __getitem__(
        self,
        item: slice | int | bool | ArrayLike,
    ) -> Self | Series | Quantity:
        """Get a slice of this SpectralVariance."""
        # disable slicing bins
        if isinstance(item, tuple) and not null_slice(item[1]):
            msg = "cannot slice SpectralVariance across bins"
            raise NotImplementedError(msg)
        return super().__getitem__(item)

    @classmethod
    def from_spectrogram(
        cls,
        *spectrograms: Spectrogram,
        bins: ArrayLike1D | None = None,
        low: float | None = None,
        high: float | None = None,
        nbins: int = 500,
        log: bool = False,
        norm: bool = False,
        density: bool = False,
    ) -> Self:
        """Calculate a new `SpectralVariance` from a Spectrogram.

        Parameters
        ----------
        *spectrograms : `~gwpy.spectrogram.Spectrogram`
            Input `Spectrogram` data.

        bins : `~numpy.ndarray`, optional
            Array of histogram bin edges, including the rightmost edge.

        low : `float`, optional
            Left edge of lowest amplitude bin, only read if ``bins`` is not given.

        high : `float`, optional
            Right edge of highest amplitude bin, only read if ``bins`` is not given.

        nbins : `int`, optional
            Number of bins to generate, only read if ``bins`` is not given,
            default: ``500``.

        log : `bool`, optional
            Calculate amplitude bins over a logarithmic scale,
            only read if ``bins`` is not given, default: `False`.

        norm : `bool`, optional
            Normalise bin counts to a unit sum, default: `False`.

        density : `bool`, optional
            Normalise bin counts to a unit integral, default: `False`.

        Returns
        -------
        specvar : `SpectralVariance`
            2D-array of spectral frequency-amplitude counts.

        See Also
        --------
        numpy.histogram
            The histogram function.
        """
        # parse args and kwargs
        if not spectrograms:
            msg = "Must give at least one Spectrogram"
            raise ValueError(msg)
        if norm and density:
            msg = "Cannot give both norm=True and density=True, please pick one"
            raise ValueError(msg)

        # get data and bins
        spectrogram = spectrograms[0]
        data = numpy.vstack([s.value for s in spectrograms])
        if bins is None:
            bins = _default_bins(
                data,
                nbins,
                low,
                high,
                spectrogram.unit,
                log=log,
            )
        else:
            bins = numpy.asarray(bins)
        nbins = bins.size - 1
        qbins = bins * spectrogram.unit

        # loop over frequencies
        out = numpy.zeros((data.shape[1], nbins))
        for i in range(data.shape[1]):
            out[i, :], bins = numpy.histogram(
                data[:, i],
                bins,
                density=density,
            )
            if norm and out[i, :].sum():  # normalise
                out[i, :] /= out[i, :].sum()

        # create and return SpectralVariance
        name = f"{spectrogram.name} variance"
        return cls(
            out,
            qbins,
            epoch=spectrogram.epoch,
            name=name,
            channel=spectrogram.channel,
            f0=spectrogram.f0,
            df=spectrogram.df,
        )

    def percentile(self, percentile: float) -> FrequencySeries:
        """Calculate a given spectral percentile for this `SpectralVariance`.

        Parameters
        ----------
        percentile : `float`
            Percentile (0 - 100) of the bins to compute.

        Returns
        -------
        spectrum : `~gwpy.frequencyseries.FrequencySeries`
            The given percentile `FrequencySeries` calculated from this
            `SpectralVariance`.
        """
        rows, _columns = self.shape
        out = numpy.zeros(rows)
        # Loop over frequencies
        for i in range(rows):
            # Calculate cumulative sum for array
            cumsumvals = numpy.cumsum(self.value[i, :])

            # Find value nearest requested percentile
            minindex = numpy.abs(cumsumvals - percentile).argmin()
            val = self.bins[minindex]
            out[i] = val

        name = f"{self.name} {percentile}% percentile"
        return FrequencySeries(
            out,
            epoch=self.epoch,
            channel=self.channel,
            frequencies=self.bins[:-1],
            name=name,
        )

    def plot(  # type: ignore[override]
        self,
        xscale: str = "log",
        method: str = "pcolormesh",
        **kwargs,
    ) -> Plot:
        """Plot the data for this SpectralVariance."""
        if method == "imshow":
            msg = (
                f"plotting a {type(self).__name__} with {method}() "
                "is not supported"
            )
            raise TypeError(msg)
        bins = self.bins.value
        if (
            numpy.all(bins > 0)
            and numpy.allclose(numpy.diff(numpy.log10(bins), n=2), 0)
        ):
            kwargs.setdefault("yscale", "log")
        kwargs.update(method=method, xscale=xscale)
        return super().plot(**kwargs)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""`Array2D` - a two-dimensional, indexed array."""

from __future__ import annotations

import contextlib
from typing import (
    TYPE_CHECKING,
    cast,
    overload,
)
from warnings import warn

from astropy.units import (
    Quantity,
    Unit,
)

from ..io.registry import UnifiedReadWriteMethod
from . import sliceutils
from .connect import (
    Array2DRead,
    Array2DWrite,
)
from .index import Index
from .series import Series

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        ClassVar,
        Self,
    )

    from astropy.units import UnitBase

    from ..plot import Plot
    from ..segments import Segment
    from ..typing import QuantityLike
    from .sliceutils import SliceLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class Array2D(Series):
    """A two-dimensional array with metadata.

    Parameters
    ----------
    value : array-like
        Input data array.

    unit : `~astropy.units.Unit`, optional
        Physical unit of these data.

    x0 : `float`, `~astropy.units.Quantity`, optional
        The starting value for the x-axis of this array.

    dx : `float`, `~astropy.units.Quantity`, optional
        The step size for the x-axis of this array.

    xindex : `array-like`
        The complete array of x-axis values for this array. This argument
        takes precedence over `x0` and `dx` so should be
        given in place of these if relevant, not alongside.

    xunit : `~astropy.units.Unit`, optional
        The unit of the x-axis coordinates. If not given explicitly, it will be
        taken from any of `dx`, `x0`, or `xindex`, or set to a boring default.

    y0 : `float`, `~astropy.units.Quantity`, optional
        The starting value for the y-axis of this array.

    dy : `float`, `~astropy.units.Quantity`, optional
        The step size for the y-axis of this array.

    yindex : `array-like`
        The complete array of y-axis values for this array. This argument
        takes precedence over `y0` and `dy` so should be
        given in place of these if relevant, not alongside.

    yunit : `~astropy.units.Unit`, optional
        The unit of the y-axis coordinates. If not given explicitly, it will be
        taken from any of `dy`, `y0`, or `yindex`, or set to a boring default.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional
        Choose to copy the input data to new memory.

    subok : `bool`, optional
        Allow passing of sub-classes by the array generator.

    Returns
    -------
    array : `Array`
        a new array, with a view of the data, and all associated metadata
    """

    _metadata_slots: ClassVar[tuple[str, ...]] = (
        *Series._metadata_slots,
        "y0",
        "dy",
        "yindex",
    )

    #: The default unit for the X-axis index.
    _default_xunit: ClassVar[UnitBase] = Unit("")

    #: The default unit for the Y-axis index.
    _default_yunit: ClassVar[UnitBase] = Unit("")

    #: The class used for viewing a row of this array.
    _rowclass: ClassVar[type[Series]] = Series

    #: The class used for viewing a column of this array.
    _columnclass: ClassVar[type[Series]] = Series

    #: The number of dimensions of this array.
    _ndim: ClassVar[int] = 2

    def __new__(
        cls,
        data: QuantityLike,
        unit: UnitBase | str | None = None,
        x0: QuantityLike | None = None,
        dx: QuantityLike | None = None,
        xindex: QuantityLike | None = None,
        xunit: UnitBase | str | None = None,
        y0: QuantityLike | None = None,
        dy: QuantityLike | None = None,
        yindex: QuantityLike | None = None,
        yunit: UnitBase | str | None = None,
        **kwargs,
    ) -> Self:
        """Define a new `Array2D`."""
        # create new object
        new = super().__new__(
            cls,
            data,
            unit=unit,
            xindex=xindex,
            xunit=xunit,
            x0=x0,
            dx=dx,
            **kwargs,
        )

        # set y-axis metadata from yindex
        if yindex is not None:
            # warn about duplicate settings
            if dy is not None:
                warn(
                    f"yindex was given to {cls.__name__}(), dy will be ignored",
                    stacklevel=2,
                )
            if y0 is not None:
                warn(
                    f"yindex was given to {cls.__name__}(), y0 will be ignored",
                    stacklevel=2,
                )
            # get unit
            if yunit is None and isinstance(yindex, Quantity):
                yunit = yindex.unit
            elif yunit is None:
                yunit = cls._default_yunit
            new.yindex = Quantity(yindex, unit=yunit)
        # or from y0 and dy
        else:
            if yunit is None and isinstance(dy, Quantity):
                yunit = dy.unit
            elif yunit is None and isinstance(y0, Quantity):
                yunit = y0.unit
            elif yunit is None:
                yunit = cls._default_yunit
            if dy is not None:
                new.dy = Quantity(dy, yunit)
            if y0 is not None:
                new.y0 = Quantity(y0, yunit)

        return new

    @overload
    def __getitem__(self, item: tuple[int, int]) -> Quantity: ...
    @overload
    def __getitem__(self, item: int) -> Series: ...
    @overload
    def __getitem__(self, item: SliceLike | tuple[SliceLike, ...]) -> Self: ...

    # rebuild getitem to handle complex slicing
    def __getitem__(  # ty: ignore[invalid-method-override]
        self,
        item: tuple[int, int] | int | SliceLike | tuple[SliceLike, ...],
    ) -> Self | Series | Quantity:
        """Get an item or a slice from this `Array2D."""
        new = super().__getitem__(item)

        # Slice axis 1 metadata
        colslice, rowslice = sliceutils.format_nd_slice(item, self.ndim)

        # Column slice
        if new.ndim == 1 and isinstance(colslice, int):
            new = cast("Series", new.view(self._columnclass))
            del new.xindex
            new.__metadata_finalize__(self)
            sliceutils.slice_axis_attributes(self, "y", new, "x", rowslice)

        # Row slice
        elif new.ndim == 1:
            new = cast("Series", new.view(self._rowclass))

        # Slice axis 1 for Array2D
        # (Series.__getitem__ will have performed column slice already)
        elif new.ndim > 1 and not sliceutils.null_slice(rowslice):
            sliceutils.slice_axis_attributes(self, "y", new, "y", rowslice)

        return new

    def __array_finalize__(self, obj: Self | None) -> None:
        """Finalize the array after a view is created."""
        super().__array_finalize__(obj)

        # Series.__array_finalize__ might set _yindex to None, so delete it
        if getattr(self, "_yindex", 0) is None:
            del self.yindex

    def __iter__(self) -> Quantity:
        """Yield the columns of this `Array2D` as `Quantity` objects."""
        # astropy Quantity.__iter__ does something fancy that we don't need
        # because we overload __getitem__
        return super(Quantity, self).__iter__()

    # -- Array2d properties ----------

    @property
    def y0(self) -> Quantity:
        """Y-axis coordinate of the first data point."""
        self._y0: Quantity
        try:
            return self._y0
        except AttributeError:
            try:
                self._y0 = self._yindex[0]
            except (AttributeError, IndexError):
                self._y0 = Quantity(0, self.yunit)
            return self._y0

    @y0.setter
    def y0(self, value: QuantityLike) -> None:
        self._update_index("y", "y0", value)

    @y0.deleter
    def y0(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._y0

    @property
    def dy(self) -> Quantity:
        """Y-axis sample separation."""
        self._dy: Quantity
        try:
            return self._dy
        except AttributeError:
            try:
                self._yindex  # noqa: B018
            except AttributeError:
                self._dy = Quantity(1, self.yunit)
            else:
                if not self.yindex.regular:
                    msg = (
                        "this series has an irregular y-axis "
                        "index, so 'dy' is not well defined"
                    )
                    raise AttributeError(msg)
                self._dy = self.yindex[1] - self.yindex[0]
            return self._dy

    @dy.setter
    def dy(self, value: QuantityLike) -> None:
        self._update_index("y", "dy", value)

    @dy.deleter
    def dy(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._dy

    @property
    def yunit(self) -> UnitBase:
        """Unit of Y-axis index."""
        try:
            return self._dy.unit
        except AttributeError:
            try:
                return self._y0.unit
            except AttributeError:
                return self._default_yunit

    @property
    def yindex(self) -> Index:
        """Positions of the data on the y-axis."""
        self._yindex: Index
        try:
            return self._yindex
        except AttributeError:
            self._yindex = Index.define(self.y0, self.dy, self.shape[1])
            return self._yindex

    @yindex.setter
    def yindex(self, index: QuantityLike) -> None:
        self._set_index("y", index)

    @yindex.deleter
    def yindex(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._yindex

    @property
    def yspan(self) -> Segment:
        """Y-axis [low, high) segment encompassed by these data.

        :type: `~gwpy.segments.Segment`
        """
        return self._index_span("y")

    @property
    def T(self) -> Self:  # noqa: N802
        """Return the transpose of this `Array2D`."""
        trans = self.value.T.view(type(self))
        trans.__array_finalize__(self)
        if hasattr(self, "_xindex"):
            trans.yindex = self.xindex.view()
        else:
            trans.y0 = self.x0
            trans.dy = self.dx
        if hasattr(self, "_yindex"):
            trans.xindex = self.yindex.view()
        else:
            trans.x0 = self.y0
            trans.dx = self.dy
        return trans

    # -- Array2D i/o -----------------

    read = UnifiedReadWriteMethod(Array2DRead)
    write = UnifiedReadWriteMethod(Array2DWrite)

    # -- Array2D methods -------------

    def _check_compatible_gwpy(
        self,
        other: Quantity,
        *,
        irregular_equal: bool = True,
    ) -> None:
        """Check whether this `Array2D` and another are compatible.

        This method checks that the Index arrays are compatible.
        """
        self._check_compatible_index(
            other,
            axis="y",
            irregular_equal=irregular_equal,
        )
        super()._check_compatible_gwpy(
            other,
            irregular_equal=irregular_equal,
        )

    def value_at(  # type: ignore[override]
        self,
        x: QuantityLike,
        y: QuantityLike,
    ) -> Quantity:
        """Return the value of this `Series` at the given `(x, y)` coordinates.

        Parameters
        ----------
        x : `float`, `~astropy.units.Quantity`
            The `xindex` value at which to search.

        y : `float`, `~astropy.units.Quantity`
            The `yindex` value at which to search.

        Returns
        -------
        z : `~astropy.units.Quantity`
            The value of this Series at the given coordinates.

        Raises
        ------
        IndexError
            If ``x`` or `y`` don't match a value on their respective index.
        """
        x = Quantity(x, self.xindex.unit).value
        y = Quantity(y, self.yindex.unit).value
        try:
            idx = (self.xindex.value == x).nonzero()[0][0]
        except IndexError as exc:
            exc.args = (f"Value {x!r} not found in array xindex",)
            raise
        try:
            idy = (self.yindex.value == y).nonzero()[0][0]
        except IndexError as exc:
            exc.args = ("Value %r not found in array yindex",)
            raise
        return self[idx, idy]

    def imshow(self, **kwargs) -> Plot:
        """Render this array on a `Plot` using `~matplotlib.pyplot.imshow`.

        Parameters
        ----------
        kwargs
            All arguments are passed to `plot`.

        See Also
        --------
        plot
            For details of plotting this object.
        """
        return self.plot(method="imshow", **kwargs)

    def pcolormesh(self, **kwargs) -> Plot:
        """Render this array on a `Plot` using `~matplotlib.pyplot.pcolormesh`.

        Parameters
        ----------
        kwargs
            All arguments are passed to `plot`.

        See Also
        --------
        plot
            For details of plotting this object.
        """
        return self.plot(method="pcolormesh", **kwargs)

    def plot(self, method: str = "imshow", **kwargs) -> Plot:
        """"Render this array on a `Plot`.

        Parameters
        ----------
        method : `str`, optional
            The `~gwpy.plot.Axes` method to call when rendering this array.

        kwargs
            All arguments are passed to `plot`.

        See Also
        --------
        gwpy.plot.Plot
            For details of plotting this object.
        """
        from ..plot import Plot

        # correct for log scales and zeros
        if kwargs.get("xscale") == "log" and self.x0.value == 0:
            kwargs.setdefault("xlim", (self.dx.value, self.xspan[1]))
        if kwargs.get("yscale") == "log" and self.y0.value == 0:
            kwargs.setdefault("ylim", (self.dy.value, self.yspan[1]))

        # make plot
        return Plot(self, method=method, **kwargs)

    # -- Array2D modifiers -----------
    # all of these try to return Quantities rather than simple numbers

    def _wrap_function(self, function: Callable, *args, **kwargs) -> Self | Quantity:
        """Wrap a function to return a new `Array2D` or `Quantity`."""
        out = super()._wrap_function(function, *args, **kwargs)

        if out.ndim != 1:
            return out

        # -- return Series
        try:
            axis = kwargs["axis"]
        except KeyError:
            axis = args[0]
        metadata = {
            "unit": out.unit,
            "channel": out.channel,
            "epoch": self.epoch,
            "name": f"{self.name} {function.__name__}",
        }

        # return Column series
        if axis == 0:
            if hasattr(self, "_yindex"):
                metadata["xindex"] = self.yindex
            else:
                metadata["x0"] = self.y0
                metadata["dx"] = self.dy
            return self._columnclass(out.value, **metadata)

        # return Row series
        if hasattr(self, "_xindex"):
            metadata["xindex"] = self.xindex
        else:
            metadata["x0"] = self.x0
            metadata["dx"] = self.dx
        return self._rowclass(out.value, **metadata)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.spectrogram`."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
    cast,
)

import pytest

from .. import SpectrogramProduct
from .base import (
    _TestFFTMixin,
    _TestImageProduct,
    _TestTimeDomainProduct,
)

if TYPE_CHECKING:
    from typing import ClassVar

    from ...spectrogram import Spectrogram

SpectrogramProductType = TypeVar("SpectrogramProductType", bound=SpectrogramProduct)


class TestSpectrogramProduct(
    _TestFFTMixin,
    _TestTimeDomainProduct[SpectrogramProductType],
    _TestImageProduct[SpectrogramProductType],
    Generic[SpectrogramProductType],
):
    """Tests for `gwpy.cli.SpectrogramProduct`."""

    TEST_CLASS: ClassVar[type[SpectrogramProduct]] = SpectrogramProduct
    ACTION = "spectrogram"

    @pytest.fixture
    @classmethod
    def dataprod(cls, prod: SpectrogramProductType) -> SpectrogramProductType:
        """Return a `SpectrogramProduct` with data."""
        cls._prod_add_data(prod)
        prod.result = prod.get_spectrogram()
        return prod

    @pytest.fixture
    @classmethod
    def plotprod(cls, dataprod: SpectrogramProductType) -> SpectrogramProductType:
        """Return a `SpectrogramProduct` with data and a plot."""
        cls._plotprod_init(dataprod)
        result = cast("Spectrogram", dataprod.result)
        dataprod.plot.gca().pcolormesh(result)
        return dataprod

    def test_get_title(self, prod: SpectrogramProductType):
        """Test `SpectrogramProduct.get_title()`."""
        assert prod.get_title() == ", ".join([
            f"fftlength={prod.args.secpfft}",
            f"overlap={prod.args.overlap}",
        ])

    def test_get_suptitle(self, prod: SpectrogramProductType):
        """Test `SpectrogramProduct.get_suptitle()`."""
        assert prod.get_suptitle() == f"Spectrogram: {prod.chan_list[0]}"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.io.gwf`."""

from pathlib import Path
from types import GeneratorType

import pytest

from ...testing.utils import (
    TEST_GWF_FILE,
    assert_segmentlist_equal,
)
from .. import gwf as io_gwf
from ..cache import file_segment

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

TEST_CHANNELS = [
    "H1:LDAS-STRAIN",
    "L1:LDAS-STRAIN",
    "V1:h_16384Hz",
]


def _backend_param(backend, *args, **kwargs):
    """Build a `pytest.param` for this GWF backend.

    This is just a convenience to apply the xfail mark for FrameL.
    """
    marks = [
        pytest.mark.requires(f"gwpy.io.gwf.{backend.lower()}"),
    ]
    if backend == "FrameL":
        marks.append(
            pytest.mark.xfail(reason="python-framel only supports getting data"),
        )
    return pytest.param(backend, *args, marks=marks, **kwargs)


parametrize_gwf_backends = pytest.mark.parametrize(
    "backend",
    [
        # backend-agnostic selection (require lalframe just to
        # skip when there are NO GWF backends installed)
        pytest.param(
            None,  # choose any backend
            marks=pytest.mark.requires("lalframe"),
            id="any",
        ),
        *map(_backend_param, io_gwf.BACKENDS),
    ],
)


def test_identify_gwf():
    """Test :func:`gwpy.io.gwf.identify_gwf`."""
    assert io_gwf.identify_gwf("read", TEST_GWF_FILE, None) is True
    with Path(TEST_GWF_FILE).open("rb") as gwff:
        assert io_gwf.identify_gwf("read", None, gwff) is True
    assert not io_gwf.identify_gwf("read", None, None)


@parametrize_gwf_backends
def test_iter_channel_names(backend):
    """Test :func:`gwpy.io.gwf.iter_channel_names`."""
    names = io_gwf.iter_channel_names(
        TEST_GWF_FILE,
        backend=backend,
    )
    assert isinstance(names, GeneratorType)
    assert list(names) == TEST_CHANNELS


@parametrize_gwf_backends
def test_get_channel_names(backend):
    """Test :func:`gwpy.io.gwf.get_channel_names`."""
    assert io_gwf.get_channel_names(
        TEST_GWF_FILE,
        backend=backend,
    ) == TEST_CHANNELS


@parametrize_gwf_backends
def test_num_channels(backend):
    """Test :func:`gwpy.io.gwf.num_channels`."""
    assert io_gwf.num_channels(
        TEST_GWF_FILE,
        backend=backend,
    ) == 3


@parametrize_gwf_backends
def test_get_channel_type(backend):
    """Test :func:`gwpy.io.gwf.get_channel_type`."""
    assert io_gwf.get_channel_type(
        TEST_CHANNELS[0],
        TEST_GWF_FILE,
        backend=backend,
    ) == "proc"
    with pytest.raises(
        ValueError,
        match=(
            "^'X1:NOT-IN_FRAME' not found in table-of-contents "
            f"for {TEST_GWF_FILE}$"
        ),
    ):
        io_gwf.get_channel_type(
            "X1:NOT-IN_FRAME",
            TEST_GWF_FILE,
            backend=backend,
        )


@parametrize_gwf_backends
@pytest.mark.parametrize(("channel", "result"), [
    (TEST_CHANNELS[0], True),
    ("X1:NOT-IN_FRAME", False),
])
def test_channel_exists(channel, result, backend):
    """Test :func:`gwpy.io.gwf.channel_exists`."""
    assert io_gwf.channel_exists(
        TEST_GWF_FILE,
        channel,
        backend=backend,
    ) is result


@parametrize_gwf_backends
def test_data_segments(backend):
    """Test :func:`gwpy.io.gwf.data_segments`."""
    assert_segmentlist_equal(
        io_gwf.data_segments(
            [TEST_GWF_FILE],
            TEST_CHANNELS[0],
            backend=backend,
        ),
        [file_segment(TEST_GWF_FILE)],
    )


@parametrize_gwf_backends
def test_data_segments_missing(backend):
    """Test :func:`gwpy.io.gwf.data_segments` with a bad channel name."""
    with pytest.warns(
        UserWarning,
        match="'X1:BAD-NAME' not found in frame",
    ):
        assert_segmentlist_equal(
            io_gwf.data_segments(
                [TEST_GWF_FILE],
                "X1:BAD-NAME",
                warn=True,
                backend=backend,
            ),
            [],
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy-plot` command line module `gwpy.cli.gwpy_plot`."""

import pytest

from .. import (
    PRODUCTS,
    gwpy_plot,
)
from .base import NDS2_CONNECTION_FIXTURE_DATA  # noqa: F401


@pytest.mark.parametrize("mode", [None, *PRODUCTS.keys()])
def test_gwpy_plot_help(mode):
    """Test the help message for `gwpy-plot` command line module."""
    args = [mode, "--help"] if mode else ["--help"]
    with pytest.raises(SystemExit) as exc:
        gwpy_plot.main(args)
    assert exc.value.code == 0


@pytest.mark.requires("nds2")
@pytest.mark.usefixtures("nds2_connection")
def test_gwpy_plot_timeseries(tmp_path):
    """Test the `gwpy-plot` command line module with a timeseries plot."""
    tmp = tmp_path / "plot.png"
    args = list(map(str, [
        "timeseries",
        "--chan", "X1:TEST-CHANNEL",
        "--start", 0,
        "--nds2-server", "nds.test.gwpy",  # don't use datafind
        "--out", tmp,
    ]))
    exitcode = gwpy_plot.main(args)

    # Check that it worked
    assert exitcode == 0

    # Check that the output file is a PNG
    with tmp.open("rb") as f:
        assert f.read(4) == b"\x89PNG"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.datafind`."""

import os
from pathlib import Path
from unittest import mock

import gwdatafind
import pytest

from ...testing.errors import pytest_skip_flaky_network
from ...testing.utils import TEST_GWF_FILE
from .. import datafind as io_datafind

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

GWDATAFIND_PATH = "gwpy.io.datafind.gwdatafind"
MOCK_ENV = {
    "VIRGODATA": "tmp",
    "GWDATAFIND_SERVER": "test:80",
}


def _mock_gwdatafind(func):
    """Decorate a function to use a mocked GWDataFind API."""
    @mock.patch.dict("os.environ", MOCK_ENV)
    @mock.patch(
        f"{GWDATAFIND_PATH}.find_types",
        mock.MagicMock(
            return_value=[Path(TEST_GWF_FILE).name.split("-")[1]],
        ),
    )
    @mock.patch(
        f"{GWDATAFIND_PATH}.find_urls",
        mock.MagicMock(return_value=[TEST_GWF_FILE]),
    )
    @mock.patch(
        f"{GWDATAFIND_PATH}.find_latest",
        mock.MagicMock(return_value=[TEST_GWF_FILE]),
    )
    @mock.patch(
        "gwpy.io.datafind.iter_channel_names",
        mock.MagicMock(return_value=["L1:LDAS-STRAIN", "H1:LDAS-STRAIN"]),
    )
    @mock.patch(
        "gwpy.io.datafind.num_channels",
        mock.MagicMock(return_value=1),
    )
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


# -- tests ---------------------------

@mock.patch.dict("os.environ", clear=True)
def test_gwdatafind_module_error():
    """Test `_gwdatafind_module()` error propagation."""
    with pytest.raises(
        RuntimeError,
        match=r"^unknown datafind configuration",
    ):
        io_datafind.find_frametype("X1:TEST")


# -- find_frametype

@_mock_gwdatafind
def test_find_frametype():
    """Test `find_frametype()`."""
    assert io_datafind.find_frametype(
        "L1:LDAS-STRAIN",
        allow_tape=True,
    ) == "HW100916"


@_mock_gwdatafind
def test_find_frametype_return_all():
    """Test `find_frametype(..., return_all=True)`."""
    assert io_datafind.find_frametype(
        "L1:LDAS-STRAIN",
        return_all=True,
    ) == ["HW100916"]


@_mock_gwdatafind
def test_find_frametype_multiple():
    """Test `find_frametype()` with list inputs."""
    assert io_datafind.find_frametype(
        ["H1:LDAS-STRAIN", "L1:LDAS-STRAIN"],
        allow_tape=True,
    ) == {
        "H1:LDAS-STRAIN": "HW100916",
        "L1:LDAS-STRAIN": "HW100916",
    }


@_mock_gwdatafind
def test_find_frametype_error_not_found():
    """Test `find_frametype()` error handing for missing channels."""
    with pytest.raises(
        ValueError,
        match=(
            r"^Cannot locate the following channel\(s\) "
            "in any known frametype:\n    X1:TEST$"
        ),
    ):
        io_datafind.find_frametype("X1:TEST", allow_tape=True)


@_mock_gwdatafind
def test_find_frametype_error_bad_channel():
    """Test `find_frametype()` error handling for bad channel names."""
    with pytest.raises(
        ValueError,
        match=(
            "^Cannot parse interferometer prefix from channel name "
            r"'bad channel name', cannot proceed with find\(\)$"
        ),
    ):
        io_datafind.find_frametype("bad channel name")


@_mock_gwdatafind
def test_find_frametype_error_files_on_tape():
    """Test `find_frametype()` error handling for ``allow_tape=False``."""
    patch = mock.patch("gwpy.io.datafind.on_tape", return_value=True)
    raises = pytest.raises(
        ValueError,
        match=r"\[files on tape have not been checked",
    )
    with patch, raises:
        io_datafind.find_frametype("X1:TEST", allow_tape=False)


# -- find_best_frametype

@_mock_gwdatafind
def test_find_best_frametype():
    """Test that `find_best_frametype` works in general."""
    assert io_datafind.find_best_frametype(
        "L1:LDAS-STRAIN",
        968654552,
        968654553,
    ) == "HW100916"


@pytest.mark.requires("lalframe")
@pytest_skip_flaky_network
@pytest.mark.skipif(
    "GWDATAFIND_SERVER" not in os.environ,
    reason="No GWDataFind server configured on this host",
)
@pytest.mark.parametrize(("channel", "expected"), [
    ("H1:ISI-GND_STS_ITMY_X_BLRMS_30M_100M.mean,s-trend", "H1_T"),
    ("H1:ISI-GND_STS_ITMY_X_BLRMS_30M_100M.mean,m-trend", "H1_M"),
])
def test_find_best_frametype_ligo_trend(channel, expected):
    """Test that `find_best_frametype` correctly matches trends.

    Currently the LIGO trend data are only available from a restricted server,
    so this test only works on LIGO computing resources.
    """
    # check that this server knows about trends and we can authenticate
    try:
        assert expected in gwdatafind.find_types("H"), (
            f"gwdatafind server doesn't know about {expected}"
        )
    except (
        AssertionError,
        RuntimeError,
    ) as exc:
        pytest.skip(str(exc))

    assert io_datafind.find_best_frametype(
        channel,
        1262276680,  # GW200105 -4s
        1262276688,  # GW200105 +4s
    ) == expected


# -- find_latest

@_mock_gwdatafind
def test_find_latest():
    """Test `find_latest()`."""
    assert io_datafind.find_latest(
        "HLV",
        "HW100916",
    ) == TEST_GWF_FILE


@mock.patch(
    f"{GWDATAFIND_PATH}.find_latest",
    mock.MagicMock(side_effect=IndexError),
)
@_mock_gwdatafind
def test_find_latest_error():
    """Test `find_latest()` error handling."""
    with pytest.raises(
        RuntimeError,
        match=r"^no files found for X-MISSING$",
    ):
        io_datafind.find_latest("X1", "MISSING")


# -- find_types

@mock.patch.dict("os.environ", MOCK_ENV)
def test_find_types():
    """Check that `find_types` works with default arguments."""
    types = ["a", "b", "c"]
    with mock.patch(f"{GWDATAFIND_PATH}.find_types", return_value=types):
        assert io_datafind.find_types(
            "X",
        ) == sorted(types)


@mock.patch.dict("os.environ", MOCK_ENV)
def test_find_types_priority():
    """Check that `find_types` prioritises trends properly."""
    types = ["L1_R", "L1_T", "L1_M"]
    with mock.patch(f"{GWDATAFIND_PATH}.find_types", return_value=types):
        assert io_datafind.find_types(
            "X",
            trend="m-trend",
        ) == ["L1_M", "L1_R", "L1_T"]


# -- utilities --------------

def test_on_tape_false():
    """Check `on_tape` works with a normal file."""
    assert io_datafind.on_tape(TEST_GWF_FILE) is False


@mock.patch("pathlib.Path.stat")
def test_on_tape_true(stat):
    """Test that `on_tape` returns `True` for a file with zero blocks."""
    stat.return_value.st_blocks = 0
    assert io_datafind.on_tape(TEST_GWF_FILE) is True


@pytest.mark.parametrize(("ftype", "trend", "priority"), [
    pytest.param("L1_HOFT_C00", None, 1, id="hoft"),
    pytest.param("H1_HOFT_C02_T1700406_v3", None, 1, id="cleaned"),
    pytest.param("H1_M", "m-trend", 0, id="m-trend"),
    pytest.param("K1_T", "s-trend", 0, id="s-trend"),
    pytest.param("K1_R", "s-trend", 5, id="raw@s-trend"),
    pytest.param("K1_M", None, 10, id="m-trend@raw"),
    pytest.param("K1_C", None, 6, id="commissioning"),
    pytest.param("SOMETHING_GRB051103", None, 10, id="lowpriority"),
    pytest.param("something else", None, 5, id="other"),
])
def test_type_priority(ftype, trend, priority):
    """Test that `_type_priority` works for various cases."""
    assert io_datafind._type_priority(
        ftype,
        trend=trend,
    )[0] == priority


# -- end-to-end tests with real data -

@pytest_skip_flaky_network
@pytest.mark.requires("lalframe", "requests_pelican")
@mock.patch.dict("os.environ", {"GWPY_CACHE": "false"})
def test_find_frametype_gwosc_osdf(caplog):
    """Test `find_frametype` with GWOSC data over OSDF."""
    caplog.set_level("DEBUG", logger="gwpy.io.datafind")
    try:
        assert io_datafind.find_frametype(
            "H1:GWOSC-4KHZ_R1_STRAIN",
            host="datafind.gwosc.org",
            urltype="osdf",
            gpstime=1269362688,
            frametype_match="_4KHZ_",
        ) == "H1_GWOSC_O3b_4KHZ_R1"
    except ValueError:
        # Astropy < 7.1 doesn't handle TimeoutError well, so check for
        # network-related issues in the logs and skip the test if so.
        for record in caplog.records:
            if "Failed to download file" in (msg := record.getMessage()):
                pytest.skip(msg)
        raise



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for the core types.

These methods are designed for internal use only.
"""

from __future__ import annotations

from numbers import Integral
from typing import (
    TYPE_CHECKING,
    overload,
)

import numpy

if TYPE_CHECKING:
    from typing import (
        Literal,
        TypeAlias,
    )

    from numpy.typing import ArrayLike

    from . import Array

    SliceLike: TypeAlias = int | bool | list | slice | ArrayLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def format_nd_slice(
    item: SliceLike | tuple[SliceLike, ...],
    ndim: int,
) -> tuple[SliceLike | None, ...]:
    """Preformat a getitem argument as an N-tuple.

    Parameters
    ----------
    item : `object`, or a `tuple` of `object` instances
        The item to format as a tuple. If a tuple, it must be of length
        less than or equal to ``ndim``. If an object, it will be converted
        to a tuple of length ``ndim`` with the object as the first element.

    ndim : `int`
        The number of dimensions to format the item for.

    Returns
    -------
    item : `tuple`
        A tuple of length ``ndim`` with the item as the first element,
        and padded with `None` values if necessary.
    """
    if not isinstance(item, tuple):
        item = (item,)
    # pad the tuple with `None` until it's ``ndim`` elements
    return item[:ndim] + (None,) * (ndim - len(item))


def slice_axis_attributes(
    old: Array,
    oldaxis: Literal["x", "y", "z"],
    new: Array,
    newaxis: Literal["x", "y", "z"],
    slice_: SliceLike,
) -> Array:
    """Set axis metadata for ``new`` by slicing an axis of ``old``.

    This is primarily for internal use in slice functions (__getitem__)

    Parameters
    ----------
    old : `Array`
        Array being sliced.

    oldaxis : ``'x'`` or ``'y'``
        The axis to slice.

    new : `Array`
        Product of slice.

    newaxis : ``'x'`` or ``'y'``
        The target axis.

    slice_ : `slice`, `numpy.ndarray`
        The slice to apply to old (or an index array).

    See Also
    --------
    Series.__getitem__
    Array2D.__getitem__
    """
    slice_ = as_slice(slice_)

    # attribute names
    index = "{}index".format
    origin = "{}0".format
    delta = "d{}".format

    # if array has an index set already, use it
    if hasattr(old, f"_{oldaxis}index"):
        setattr(
            new,
            index(newaxis),
            getattr(old, index(oldaxis))[slice_],
        )
        return new

    # otherwise if using a slice, use origin and delta properties
    if isinstance(slice_, slice) or not numpy.sum(slice_):
        if isinstance(slice_, slice):
            offset = slice_.start or 0
            step = slice_.step or 1
        else:  # empty ndarray slice (so just set attributes)
            offset = 0
            step = 1

        dx = getattr(old, delta(oldaxis))
        x0 = getattr(old, origin(oldaxis))

        # set new.x0 / new.y0
        setattr(new, origin(newaxis), x0 + offset * dx)

        # set new.dx / new.dy
        setattr(new, delta(newaxis), dx * step)

    # otherwise slice with an index array
    else:
        setattr(new, index(newaxis), getattr(old, index(oldaxis))[slice_])

    return new


def null_slice(slice_: SliceLike) -> bool:
    """Return `True` if a slice will have no affect."""
    try:
        slice_ = as_slice(slice_)
    except TypeError:
        return False

    # trivial array slice
    if (
        isinstance(slice_, numpy.ndarray)
        and slice_.dtype == bool
        and slice_.all()
    ):
        return True

    # trivial slice object
    return (
        isinstance(slice_, slice)
        and slice_ in (
            slice(None, None, None),
            slice(0, None, 1),
        )
    )


@overload
def as_slice(slice_: SliceLike) -> slice | numpy.ndarray:
    ...


@overload
def as_slice(slice_: tuple[SliceLike, ...]) -> tuple[slice | numpy.ndarray, ...]:
    ...


def as_slice(
    slice_: None | SliceLike | tuple[SliceLike, ...],
) -> slice | numpy.ndarray | tuple[slice | numpy.ndarray, ...]:
    """Convert an object to a slice, or tuple of slices, if possible.

    slice_ : `int`, `list` `numpy.ndarray`, `None`
        The input to cast to a `slice_`.

    Returns
    -------
    slice : `slice` or `tuple` of `slice`
        The input cast into a single slice, or a tuple of slices.
    """
    if slice_ is None or isinstance(slice_, Integral | numpy.integer):
        return slice(0, None, 1)

    if isinstance(slice_, list):
        slice_ = numpy.array(slice_)

    if isinstance(slice_, slice | numpy.ndarray):
        return slice_

    if isinstance(slice_, tuple):
        return tuple(map(as_slice, slice_))

    msg = f"cannot format '{slice_}' as slice"
    raise TypeError(msg)



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Coherence spectrogram."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .spectrogram import SpectrogramProduct

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from logging import Logger
    from typing import ClassVar

    from ..spectrogram import Spectrogram

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

logger = logging.getLogger(__name__)


class CoherencegramProduct(SpectrogramProduct):
    """Plot the coherence-spectrogram comparing two time series."""

    DEFAULT_CMAP: ClassVar[str] = "plasma"
    MIN_DATASETS: ClassVar[int] = 2
    MAX_DATASETS: ClassVar[int] = 2
    action: ClassVar[str] = "coherencegram"

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Create a new `CoherencegramProduct`."""
        super().__init__(args, logger=logger)
        self.ref_chan = self.args.ref or self.chan_list[0]
        # deal with channel type (e.g. m-trend)
        self.ref_chan = self.ref_chan.split(",", maxsplit=1)[0]

    @classmethod
    def arg_channels(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Configure `~argparse.ArgumentParser` arguments for channels."""
        group = super().arg_channels(parser)
        group.add_argument(
            "--ref",
            help="Reference channel against which others will be compared",
        )
        return group

    def _finalize_arguments(self, args: Namespace) -> None:
        """Finalise arguments with defaults."""
        if args.color_scale is None:
            args.color_scale = "linear"
        if args.color_scale == "linear":
            if args.imin is None:
                args.imin = 0.0
            if args.imax is None:
                args.imax = 1.0
        return super()._finalize_arguments(args)

    def get_ylabel(self) -> str:
        """Text for y-axis label."""
        return "Frequency (Hz)"

    def get_suptitle(self) -> str:
        """Start of default super title, first channel is appended to it."""
        a, b = self.chan_list
        return f"Coherence spectrogram: {a} vs {b}"

    def get_color_label(self) -> str:
        """Return the default colorbar label."""
        if self.args.norm:
            return f"Normalized to {self.args.norm}"
        return "Coherence"

    def get_stride(self) -> float:
        """Calculate the stride for the coherencegram."""
        fftlength = float(self.args.secpfft)
        overlap = self.args.overlap  # fractional overlap
        return max(
            self.duration / (self.width * 0.8),
            fftlength * (1 + (1 - overlap) * 32),
            fftlength * 2,
        )

    def get_spectrogram(self) -> Spectrogram:
        """Calculate the `Spectrogram` to be plotted."""
        args = self.args
        fftlength = float(args.secpfft)
        overlap = args.overlap  # fractional overlap
        stride = self.get_stride()
        self.logger.debug(
            "Calculating coherence spectrogram, secpfft: %s, overlap: %s",
            fftlength,
            overlap,
        )

        if overlap is not None:  # overlap in seconds
            overlap *= fftlength

        if self.timeseries[0].name == self.ref_chan:
            ref = 0
            other = 1
        else:
            ref = 1
            other = 0
        return self.timeseries[ref].coherence_spectrogram(
            self.timeseries[other],
            stride,
            fftlength=fftlength,
            overlap=overlap,
            window=args.window,
        )



# Copyright (c) 2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom default figure configuration."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from matplotlib import (
    RcParams,
    rc_params as mpl_rc_params,
    rcParams,
)

from ..utils.env import bool_env
from . import tex

if TYPE_CHECKING:
    from matplotlib.figure import SubplotParams

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# record matplotlib's original rcParams
MPL_RCPARAMS = mpl_rc_params()

# record the LaTeX preamble
PREAMBLE = rcParams.get("text.latex.preamble", "") + os.linesep.join(tex.MACROS)

# -- custom rc -----------------------

# set default params
GWPY_RCPARAMS = RcParams(**{
    # axes boundary colours
    "axes.edgecolor": "gray",
    # grid
    "axes.grid": True,
    "axes.axisbelow": False,
    "grid.linewidth": .5,
    # ticks
    "axes.formatter.limits": (-3, 4),
    "axes.formatter.use_mathtext": True,
    "axes.labelpad": 5,
    # fonts
    "axes.titlesize": "large",
    "axes.labelsize": "large",
    "font.family": ["sans-serif"],
    "font.sans-serif": [
        "FreeSans",
        "Helvetica Neue",
        "Helvetica",
        "Arial",
    ] + rcParams["font.sans-serif"],
    "font.size": 12,
    # legend (revert to mpl 1.5 formatting in parts)
    "legend.edgecolor": "inherit",
    "legend.numpoints": 2,
    "legend.handlelength": 1,
    "legend.fancybox": False,
})

# set latex options
GWPY_TEX_RCPARAMS = RcParams(**{
    # use latex styling
    "text.usetex": True,
    "text.latex.preamble": PREAMBLE,
    # use bigger font for labels (since the font is good)
    "font.family": ["serif"],
    "font.size": 16,
    # don't use mathtext for offset
    "axes.formatter.use_mathtext": False,
})


def rc_params(*, usetex: bool | None = None) -> RcParams:
    """Return a new `matplotlib.RcParams` with updated GWpy parameters.

    The updated parameters are globally stored as
    `gwpy.plot.rc.GWPY_RCPARAMS`, with the updated TeX parameters as
    `gwpy.plot.rc.GWPY_TEX_RCPARAMS`.

    .. note::

       This function doesn't apply the new `RcParams` in any way, just
       creates something that can be used to set `matplotlib.rcParams`.

    Parameters
    ----------
    usetex : `bool`, `None`
        Value to set for `text.usetex`; if `None` (default) determine
        automatically using the ``GWPY_USETEX`` environment variable,
        and whether `tex` is available on the system.
        If `True` is given (or determined) a number of other parameters
        are updated to improve TeX formatting.

    Returns
    -------
    rcparams: `matplotlib.RcParams`
        The new parameters set.

    Examples
    --------
    >>> import matplotlib
    >>> from gwpy.plot.rc import rc_params as gwpy_rc_params()
    >>> matplotlib.rcParams.update(gwpy_rc_params(usetex=False))
    """
    # if user didn't specify to use tex or not, guess based on
    # the `GWPY_USETEX` environment variable, or whether tex is
    # installed at all.
    if usetex is None:
        usetex = bool_env(
            "GWPY_USETEX",
            default=rcParams["text.usetex"] or tex.has_tex(),
        )

    # build RcParams from matplotlib.rcParams with GWpy extras
    rcp = GWPY_RCPARAMS.copy()
    if usetex:
        rcp.update(GWPY_TEX_RCPARAMS)
    return rcp


# -- dynamic subplot positioning -----

SUBPLOT_WIDTH = {
    6.4: (.1875, .87),
    8.: (.15, .85),
    12.: (.1, .90),
}
SUBPLOT_HEIGHT = {
    3.: (.25, .83),
    4.: (.2, .85),
    4.8: (.16, .88),
    5: (.15, .89),
    6.: (.13, .9),
    8.: (.1, .93),
}


def get_subplot_params(figsize: tuple[float, float]) -> SubplotParams:
    """Return sensible default `SubplotParams` for a figure of the given size.

    Parameters
    ----------
    figsize : `tuple` of `float`
         The ``(width, height)`` figure size (inches).

    Returns
    -------
    params : `~matplotlib.figure.SubplotParams`
        Formatted set of subplot parameters.
    """
    from matplotlib.figure import SubplotParams

    left: float | None
    right: float | None
    bottom: float | None
    top: float | None

    width, height = figsize
    try:
        left, right = SUBPLOT_WIDTH[width]
    except KeyError:
        left = right = None
    try:
        bottom, top = SUBPLOT_HEIGHT[height]
    except KeyError:
        bottom = top = None
    return SubplotParams(
        left=left,
        right=right,
        bottom=bottom,
        top=top,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for plotting data with matplotlib and LAL."""

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from matplotlib import rcParams

if TYPE_CHECKING:
    from collections.abs import (
        Iterable,
        Iterator,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# groups of input parameters (for passing to Plot())
FIGURE_PARAMS: list[str] = [
    "dpi",
    "figsize",
]
AXES_PARAMS: list[str] = [
    # standard options
    "projection",
    "title",
    # X-axis params
    "sharex",
    "xlabel",
    "xlim",
    "xscale",
    # Y-axis params
    "sharey",
    "ylabel",
    "ylim",
    "yscale",
    # special GWpy extras
    "epoch",
    "insetlabels",
]


def color_cycle(colors: Iterable[str] | None = None) -> Iterator[str]:
    """Return an infinite iterator of the given (or default) colors.

    Parameters
    ----------
    colors : `list` of `str`, optional
        The colours to iterate.
        Default is `None` to use Matplotlib's default set of colours.

    Returns
    -------
    colors : iterator of `str`
        The new iterator that yields colour strings infinitely.
    """
    if colors:
        return itertools.cycle(colors)
    return itertools.cycle(p["color"] for p in rcParams["axes.prop_cycle"])


def marker_cycle(markers: Iterable[str] | None = None) -> Iterator[str]:
    """Return an infinite iterator of the given (or default) markers.

    Parameters
    ----------
    markers : `list` of `str`, optional
        The markers to iterate.
        Default is `None` to use a standard set of markers.

    Returns
    -------
    markers : iterator of `str`
        The new iterator that yields marker strings infinitely.
    """
    if markers:
        return itertools.cycle(markers)
    return itertools.cycle(("o", "x", "+", "^", "D", "H", "1"))



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.sql`."""

from ...testing.errors import pytest_skip_flaky_network
from .. import EventTable


@pytest_skip_flaky_network
def test_fetch_gwosc():
    """Test fetching a GWOSC event catalog."""
    table = EventTable.fetch(
        "GWTC-1-confident",
        source="gwosc",
    )
    assert len(table)
    assert {
        "mass_1_source",
        "luminosity_distance",
        "chi_eff",
    }.intersection(table.colnames)

    # check unit parsing worked
    assert table["luminosity_distance"].unit == "Mpc"


@pytest_skip_flaky_network
def test_fetch_open_data_kwargs():
    """Test fetching a GWOSC event catalog with ``where`` and ``columns``."""
    table = EventTable.fetch_open_data(
        "GWTC-1-confident",
        where="mass_1_source < 5",
        columns=[
            "name",
            "mass_1_source",
            "mass_2_source",
            "luminosity_distance",
        ],
    )
    assert len(table) == 1
    assert table[0]["name"] == "GW170817-v3"
    assert set(table.colnames) == {
        "name",
        "mass_1_source",
        "mass_2_source",
        "luminosity_distance",
    }



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from GWF FrEvent structures into a Table."""

from __future__ import annotations

from typing import TYPE_CHECKING

from astropy.table import Table

from ...io import gwf as io_gwf
from ...io.utils import FileLike
from ...time import LIGOTimeGPS
from .. import EventTable
from ..filter import parse_column_filters

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from typing import IO

    from LDASTools.frameCPP import FrEvent

    from ..filter import FilterSpec

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

def _columns_from_frevent(frevent: FrEvent) -> list[str]:
    """Get list of column names from frevent."""
    params = dict(frevent.GetParam())
    return [
        "time",
        "amplitude",
        "probability",
        "timeBefore",
        "timeAfter",
        "comment",
        *list(params.keys()),
    ]


def _row_from_frevent(
    frevent: FrEvent,
    columns: Iterable[str],
    where: Iterable[FilterSpec],
) -> list[str | float] | None:
    """Generate a table row from an FrEvent.

    Filtering (``where``) is done here, rather than in the table reader,
    to enable filtering on columns that aren't being returned.

    Returns `None` if this event doesn't match the ``where`` conditions.
    """
    # read params
    params = dict(frevent.GetParam())
    params["time"] = float(LIGOTimeGPS(*frevent.GetGTime()))
    params["amplitude"] = frevent.GetAmplitude()
    params["probability"] = frevent.GetProbability()
    params["timeBefore"] = frevent.GetTimeBefore()
    params["timeAfter"] = frevent.GetTimeAfter()
    params["comment"] = frevent.GetComment()

    # filter
    if not all(op_(params[c], t) for c, op_, t in where):
        return None

    # return event as list
    return [params[c] for c in columns]


def table_from_gwf(
    filename: str | Path | IO,
    name: str,
    columns: Iterable[str] | None = None,
    where: str | list[str] | None = None,
) -> Table:
    """Read a Table from FrEvent structures in a GWF file (or files).

    This method requires |LDAStools.frameCPP|_.

    Parameters
    ----------
    filename : `str`, `pathlib.Path`, `file`
        The path of GWF file to read.

    name : `str`
        The name associated with the `FrEvent` structures.

    columns : `list` of `str`
        List of column names to read.

    where : `str`, `list` of `str`
        One or more filter condition strings to apply, e.g. ``'snr>6'``.

    Returns
    -------
    Table
        A table of rows read from the GWF input.
    """
    gwf_framecpp = io_gwf.import_backend("frameCPP")

    # open frame file
    if isinstance(filename, FileLike):
        filename = filename.name
    stream = gwf_framecpp.open_gwf(filename)

    # parse where conditions and map to column indices
    if where is None:
        where = []
    conditions = parse_column_filters(where)

    # read events row by row
    data = []
    i = 0
    while True:
        try:
            frevent = stream.ReadFrEvent(i, name)
        except IndexError:
            break
        i += 1
        # read first event to get column names
        if columns is None:
            columns = _columns_from_frevent(frevent)
        # read row with filter
        row = _row_from_frevent(frevent, columns, conditions)
        if row is not None:  # if passed condition
            data.append(row)

    return Table(rows=data, names=columns)


# -- write ---------------------------

def table_to_gwf(
    table: Table,
    filename: str | Path,
    name: str,
    **kwargs,
) -> None:
    """Create a new `~frameCPP.FrameH` and fill it with data.

    Parameters
    ----------
    table : `~astropy.table.Table`
        the data to write

    filename : `str`
        the name of the file to write into

    name: `str`, optional
        The name to give the ``FrameH`` object **and** each ``FrEvent``.

    kwargs
        Other keyword arguments are passed to the GWF creator and writer
        functions (see below).

    See Also
    --------
    gwpy.io.gwf.create_frame
        For details of how the GWF ``FrameH`` structure is created and
        any valid keyword arguments.

    gwpy.io.gwf.write_frames
        For details of the GWF ``FrEvent`` structures are created, and
        any valid keyword arguments.
    """
    from LDAStools.frameCPP import (
        FrEvent,
        GPSTime,
    )

    gwf_framecpp = io_gwf.import_backend("frameCPP")

    # create frame
    write_kw = {
        key: kwargs.pop(key)
        for key in (
            "compression",
            "compression_level",
        ) if key in kwargs
    }
    frame = gwf_framecpp.create_frame(
        name=name,
        **kwargs,
    )

    # append row by row
    names = table.dtype.names
    for row in table:
        rowd = {n: row[n] for n in names}
        gps = LIGOTimeGPS(rowd.pop("time", 0))
        frame.AppendFrEvent(FrEvent(
            str(name),
            str(rowd.pop("comment", "")),
            str(rowd.pop("inputs", "")),
            GPSTime(gps.gpsSeconds, gps.gpsNanoSeconds),
            float(rowd.pop("timeBefore", 0)),
            float(rowd.pop("timeAfter", 0)),
            int(rowd.pop("eventStatus", 0)),
            float(rowd.pop("amplitude", 0)),
            float(rowd.pop("probability", -1)),
            str(rowd.pop("statistics", "")),
            list(rowd.items()),  # remaining params as tuple
        ))

    # write frame to file
    gwf_framecpp.write_frames(
        filename,
        [frame],
        **write_kw,
    )


# -- registration --------------------

for klass in (Table, EventTable):
    # read
    klass.read.registry.register_reader(
        "gwf",
        klass,
        table_from_gwf,
    )
    # write
    klass.write.registry.register_writer(
        "gwf",
        klass,
        table_to_gwf,
    )
    # identify
    klass.read.registry.register_identifier(
        "gwf",
        klass,
        io_gwf.identify_gwf,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Support for plotting with units."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

from astropy.units.format import LatexInline

if TYPE_CHECKING:
    from astropy.units import UnitBase


class LatexInlineDimensional(LatexInline):
    r"""Custom LaTeX formatter that includes physical type (if available).

    Mainly for auto-labelling `Axes` in matplotlib figures.

    Examples
    --------
    The built-in astropy ``latex_inline`` formatter gives this:

    >>> Unit('m/s').to_string(format='latex_inline')
    '$\mathrm{m\,s^{-1}}$'

    This custom 'dimensional' formatter gives:

    >>> Unit('m/s').to_string(format='latex_inline_dimensional')
    'Speed/Velocity [$\mathrm{m\,s^{-1}}$]'
    """

    name = "latex_inline_dimensional"

    @classmethod
    @wraps(LatexInline.to_string)
    def to_string(
        cls,
        unit: UnitBase,
        **kwargs,
    ) -> str:
        """Output this unit in LaTeX format with dimensions."""
        if unit.physical_type == "dimensionless":
            return "Dimensionless"

        u = f"[{super().to_string(unit, **kwargs)}]"

        if (
            unit.physical_type is None
            or unit.physical_type == "unknown"
        ):
            return u

        # format physical type of unit for LaTeX
        ptype = str(unit.physical_type).title().replace("_", r"\_")
        # looks like '<Physical type> [<unit>]'
        return f"{ptype} {u}"



# Copyright (c) 2017 Louisiana State University
#               2017-2022 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extend :mod:`astropy.table` with the `EventTable`."""

from __future__ import annotations

from functools import wraps
from math import ceil
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from astropy.table import (
    Table,
)
from gwosc.api import DEFAULT_URL as DEFAULT_GWOSC_URL

from ..io.registry import (
    UnifiedReadWriteMethod,
    inherit_unified_io,
)
from ..time import (
    LIGOTimeGPSLike,
    to_gps,
)
from .connect import (
    EventTableFetch,
    EventTableRead,
    EventTableWrite,
)
from .filter import (
    filter_table,
    parse_operator,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Sequence,
    )
    from typing import (
        ParamSpec,
        Self,
        TypeVar,
    )

    from astropy.table import Column

    from ..plot import Plot
    from ..time import SupportsToGps
    from ..timeseries import (
        TimeSeries,
        TimeSeriesDict,
    )
    from .filter import FilterLike

    # ParamSpec for decorators
    P = ParamSpec("P")
    R = TypeVar("R")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

TIME_LIKE_COLUMN_NAMES = [
    "time",  # standard
    "gps",  # GWOSC catalogues
    "peakGPS",  # gravityspy
]


# -- utilities -----------------------

def _rates_preprocess(func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> R:
        self: EventTable = args[0]  # type: ignore[assignment]
        timecolumn = kwargs.get("timecolumn")
        start = cast("SupportsToGps | None", kwargs.get("start"))
        end = cast("SupportsToGps | None", kwargs.get("end"))

        # get timecolumn if we are going to need it
        if (
            (timecolumn is None and (start is None or end is None))
            or not self.colnames
        ):
            try:
                kwargs["timecolumn"] = self._get_time_column()
            except ValueError as exc:
                exc.args = (f"{exc.args[0]}, please give `timecolumn` keyword",)
                raise
        # otherwise use anything (it doesn't matter)
        kwargs.setdefault("timecolumn", self.colnames[0])

        # set start and end
        times = self[kwargs["timecolumn"]]
        if start is None:
            start = times.min()
        else:
            start = to_gps(start)
        if end is None:
            end = times.max()
        else:
            end = to_gps(end)
        kwargs["start"] = start
        kwargs["end"] = end

        return func(*args, **kwargs)
    return wrapped_func


# -- Table ---------------------------

@inherit_unified_io
class EventTable(Table):
    """A container for a table of events.

    This object expands the default :class:`~astropy.table.Table`
    with extra read/write formats, and methods to perform filtering,
    rate calculations, and visualisation.

    See Also
    --------
    astropy.table.Table
        for details on parameters for creating an `EventTable`
    """

    # -- utilities -------------------

    def _is_time_column(self, name: str) -> bool:
        """Return `True` if a column in this table represents 'time'.

        This method checks the name of the column against a hardcoded list
        of time-like names, then checks the first element of the named
        columne to see if it looks like a `LIGOTimeGPS`-like object.
        """
        # if the name looks like a time column, accept that
        if name.lower() in TIME_LIKE_COLUMN_NAMES:
            return True

        # if the dtype of this column looks right, accept that
        try:
            return isinstance(self[name][0], LIGOTimeGPSLike)
        except IndexError:
            return False

    def _get_time_column(self) -> str:
        """Return the name of the 'time' column in this table.

        This method tries the following:

        - look for a single column named 'time', 'gps', or 'peakGPS'
        - look for a single column with a GPS type (e.g. `LIGOTimeGPS`)

        So, its not foolproof.

        Raises a `ValueError` if either 0 or multiple matches are found.
        """
        matches = list(filter(self._is_time_column, self.columns))
        try:
            time, = matches
        except ValueError:
            tcolnames = ", or ".join(
                ", ".join(map(repr, TIME_LIKE_COLUMN_NAMES)).rsplit(", ", 1),
            )
            msg = (
                "cannot identify time column for table, no columns "
                f"named {tcolnames}, or with a GPS dtype"
            )
            if len(matches) > 1:
                msg = msg.replace("no columns", "multiple columns")
            raise ValueError(msg) from None
        return time

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(EventTableRead)
    write = UnifiedReadWriteMethod(EventTableWrite)
    fetch = UnifiedReadWriteMethod(EventTableFetch)

    @classmethod
    def fetch_open_data(
        cls,
        catalog: str,
        columns: list[str] | None = None,
        where: str | list[str] | None = None,
        host: str = DEFAULT_GWOSC_URL,
        **kwargs,
    ) -> Self:
        """Fetch events from an open-data catalogue hosted by GWOSC.

        This is an alias for `EventTable.fetch(format='gwosc')`.

        Parameters
        ----------
        catalog : `str`
            The name of the catalog to fetch, e.g. ``'GWTC-1-confident'``.

        columns : `list` of `str`, optional
            The list of column names to read.

        where : `str`, or `list` of `str`, optional
            One or more column filters with which to downselect the
            returned table rows as they as read, e.g. ``'snr > 5'``,
            similar to a SQL ``WHERE`` statement.
            Multiple conditions should be connected by ' && ' or ' and ',
            or given a `list`, e.g. ``'mchirp < 3 && distance < 500'`` or
            ``['mchirp < 3', 'distance < 500']``

        host : `str`, optional
            The open-data host to use.

        **kwargs
            Other keyword arguments are passed to the fetch method.
        """
        return cls.fetch(
            source="gwosc",
            catalog=catalog,
            columns=columns,
            where=where,
            host=host,
            **kwargs,
        )

    # -- ligolw compatibility --------

    def get_column(self, name: str) -> Column:
        """Return the `Column` with the given name.

        This method is provided only for compatibility with the
        :class:`igwn_ligolw.ligolw.Table`.

        Parameters
        ----------
        name : `str`
            the name of the column to return

        Returns
        -------
        column : `astropy.table.Column`

        Raises
        ------
        KeyError
            if no column is found with the given name
        """
        return self[name]

    # -- extensions ------------------

    @_rates_preprocess
    def event_rate(
        self,
        stride: float,
        start: SupportsToGps | None = None,
        end: SupportsToGps | None = None,
        timecolumn: str | None = None,
    ) -> TimeSeries:
        """Calculate the rate `~gwpy.timeseries.TimeSeries` for this `Table`.

        Parameters
        ----------
        stride : `float`
            size (seconds) of each time bin

        start : `float`, `~gwpy.time.LIGOTimeGPS`, optional
            GPS start epoch of rate `~gwpy.timeseries.TimeSeries`

        end : `float`, `~gwpy.time.LIGOTimeGPS`, optional
            GPS end time of rate `~gwpy.timeseries.TimeSeries`.
            This value will be rounded up to the nearest sample if needed.

        timecolumn : `str`, optional
            name of time-column to use when binning events, attempts
            are made to guess this

        Returns
        -------
        rate : `~gwpy.timeseries.TimeSeries`
            a `TimeSeries` of events per second (Hz)

        Raises
        ------
        ValueError
            if the ``timecolumn`` cannot be guessed from the table contents
        """
        from gwpy.timeseries import TimeSeries

        # NOTE: decorator sets timecolumn, start, end to non-None values
        timecolumn = cast("str", timecolumn)
        start = cast("float", start)
        end = cast("float", end)

        times = self[timecolumn]
        if times.dtype.name == "object":  # cast to ufuncable type
            times = times.astype("longdouble", copy=False)
        nsamp = ceil((end - start) / stride)
        timebins = numpy.arange(nsamp + 1) * stride + start

        # create histogram
        return TimeSeries(
            numpy.histogram(times, bins=timebins)[0] / float(stride),
            t0=start,
            dt=stride,
            unit="Hz",
            name="Event rate",
        )

    @_rates_preprocess
    def binned_event_rates(
        self,
        stride: float,
        column: str,
        bins: Sequence[tuple[float, float]] | Sequence[float],
        operator: str | Callable[[object, object], bool] = ">=",
        start: SupportsToGps | None = None,
        end: SupportsToGps | None = None,
        timecolumn: str | None = None,
    ) -> TimeSeriesDict:
        """Calculate an event rate `~gwpy.timeseries.TimeSeriesDict`.

        Calculate the event rate over a number of bins.

        Parameters
        ----------
        stride : `float`
            size (seconds) of each time bin

        column : `str`
            name of column by which to bin.

        bins : `list`
            a list of `tuples <tuple>` marking containing bins, or a list of
            `floats <float>` defining bin edges against which an math operation
            is performed for each event.

        operator : `str`, `callable`
            one of:

            - ``'<'``, ``'<='``, ``'>'``, ``'>='``, ``'=='``, ``'!='``,
              for a standard mathematical operation,
            - ``'in'`` to use the list of bins as containing bin edges, or
            - a callable function that takes compares an event value
              against the bin value and returns a boolean.

            .. note::

               If ``bins`` is given as a list of tuples, this argument
               is ignored.

        start : `float`, `~gwpy.time.LIGOTimeGPS`, optional
            GPS start epoch of rate `~gwpy.timeseries.TimeSeries`.

        end : `float`, `~gwpy.time.LIGOTimeGPS`, optional
            GPS end time of rate `~gwpy.timeseries.TimeSeries`.
            This value will be rounded up to the nearest sample if needed.

        timecolumn : `str`, optional, default: ``time``
            name of time-column to use when binning events

        Returns
        -------
        rates : ~gwpy.timeseries.TimeSeriesDict`
            a dict of (bin, `~gwpy.timeseries.TimeSeries`) pairs describing a
            rate of events per second (Hz) for each of the bins.
        """
        from gwpy.timeseries import TimeSeriesDict

        # NOTE: decorator sets timecolumn, start, end to non-None values
        timecolumn = cast("str", timecolumn)
        start = cast("float", start)
        end = cast("float", end)

        # generate column bins
        if not bins:
            bins = [(-numpy.inf, numpy.inf)]
        if operator == "in" and not isinstance(bins[0], tuple):
            bins = cast("list[float]", bins)
            bins = [(bin_, bins[i+1]) for i, bin_ in enumerate(bins[:-1])]
        elif isinstance(operator, str):
            op_func = parse_operator(operator)
        else:
            op_func = operator

        coldata = self[column]

        # generate one TimeSeries per bin
        out = TimeSeriesDict()
        for bin_ in bins:
            if isinstance(bin_, tuple):
                keep = (coldata >= bin_[0]) & (coldata < bin_[1])
            else:
                keep = op_func(coldata, bin_)
            out[bin_] = self[keep].event_rate(
                stride,
                start=start,
                end=end,
                timecolumn=timecolumn,
            )
            out[bin_].name = " ".join((column, str(operator), str(bin_)))

        return out

    def scatter(self, x: str, y: str, **kwargs) -> Plot:
        """Make a scatter plot of column ``x`` vs column ``y``.

        Parameters
        ----------
        x : `str`
            name of column defining centre point on the X-axis

        y : `str`
            name of column defining centre point on the Y-axis

        color : `str`, optional, default:`None`
            name of column by which to color markers

        **kwargs
            any other keyword arguments, see below

        Returns
        -------
        plot : `~gwpy.plot.Plot`
            the newly created figure

        See Also
        --------
        matplotlib.pyplot.figure
            for documentation of keyword arguments used to create the
            figure
        matplotlib.figure.Figure.add_subplot
            for documentation of keyword arguments used to create the
            axes
        gwpy.plot.Axes.scatter
            for documentation of keyword arguments used to display the table
        """
        color = kwargs.pop("color", None)
        if color is not None:
            kwargs["c"] = self[color]
        return self._plot("scatter", self[x], self[y], **kwargs)

    def tile(self, x: str, y: str, w: str, h: str, **kwargs) -> Plot:
        """Make a tile plot of this table.

        Parameters
        ----------
        x : `str`
            name of column defining anchor point on the X-axis

        y : `str`
            name of column defining anchor point on the Y-axis

        w : `str`
            name of column defining extent on the X-axis (width)

        h : `str`
            name of column defining extent on the Y-axis (height)

        color : `str`, optional, default:`None`
            name of column by which to color markers

        **kwargs
            any other keyword arguments, see below

        Returns
        -------
        plot : `~gwpy.plot.Plot`
            the newly created figure

        See Also
        --------
        matplotlib.pyplot.figure
            for documentation of keyword arguments used to create the
            figure
        matplotlib.figure.Figure.add_subplot
            for documentation of keyword arguments used to create the
            axes
        gwpy.plot.Axes.tile
            for documentation of keyword arguments used to display the table
        """
        color = kwargs.pop("color", None)
        if color is not None:
            kwargs["color"] = self[color]
        return self._plot("tile", self[x], self[y], self[w], self[h], **kwargs)

    def _plot(self, method: str, *args: Column, **kwargs) -> Plot:
        from matplotlib import rcParams

        from ..plot import Plot
        from ..plot.tex import label_to_latex

        if self._is_time_column(args[0].name):
            # map X column to GPS axis
            kwargs.setdefault("figsize", (12, 6))
            kwargs.setdefault("xscale", "auto-gps")

        kwargs["method"] = method
        plot = Plot(*args, **kwargs)

        # set default labels
        ax = plot.gca()
        for axis, col in zip((ax.xaxis, ax.yaxis), args[:2], strict=False):
            if not axis.isDefault_label:
                continue
            name = col.name
            if rcParams["text.usetex"]:
                name = rf"\texttt{{{label_to_latex(col.name)}}}"
            if col.unit is not None:
                name += " [{}]".format(col.unit.to_string("latex_inline"))
            axis.set_label_text(name)
            axis.isDefault_label = True

        return plot

    def hist(self, column: str, **kwargs) -> Plot:
        """Generate a `HistogramPlot` of this `Table`.

        Parameters
        ----------
        column : `str`
            Name of the column over which to histogram data

        method : `str`, optional
            Name of `~matplotlib.axes.Axes` method to use to plot the
            histogram, default: ``'hist'``.

        **kwargs
            Any other keyword arguments, see below.

        Returns
        -------
        plot : `~gwpy.plot.Plot`
            The newly created figure.

        See Also
        --------
        matplotlib.pyplot.figure
            for documentation of keyword arguments used to create the
            figure.
        matplotlib.figure.Figure.add_subplot
            for documentation of keyword arguments used to create the
            axes.
        gwpy.plot.Axes.hist
            for documentation of keyword arguments used to display the
            histogram, if the ``method`` keyword is given, this method
            might not actually be the one used.
        """
        from ..plot import Plot
        return Plot(self[column], method="hist", **kwargs)

    def filter(self, *column_filters: FilterLike) -> Self:
        """Apply one or more column slice filters to this `EventTable`.

        Multiple column filters can be given, and will be applied
        concurrently

        Parameters
        ----------
        *column_filters : `str`, `tuple`
            a column slice filter definition, e.g. ``'snr > 10``, or
            a filter tuple definition, e.g. ``('snr', <my_func>, <arg>)``

        Notes
        -----
        See :ref:`gwpy-table-filter` for more details on using filter tuples

        Returns
        -------
        table : `EventTable`
            a new table with only those rows matching the filters

        Examples
        --------
        To filter an existing `EventTable` (``table``) to include only
        rows with ``snr`` greater than `10`, and ``frequency`` less than
        `1000`:

        >>> table.filter('snr>10', 'frequency<1000')

        Custom operations can be defined using filter tuple definitions:

        >>> from gwpy.table.filters import in_segmentlist
        >>> table.filter(('time', in_segmentlist, segs))
        """
        return filter_table(self, *column_filters)

    def cluster(
        self,
        index: str,
        rank: str,
        window: float,
    ) -> Self:
        """Cluster this `EventTable` over a given column.

        Cluster over the `index` column, maximizing over the `rank` column
        in the table.

        The clustering algorithm uses a pooling method to identify groups
        of points that are all separated in `index` by less than `window`.

        Each cluster of nearby points is replaced by the point in that cluster
        with the maximum value of `rank`.

        Parameters
        ----------
        index : `str`
            name of the column which is used to search for clusters

        rank : `str`
            name of the column to maximize over in each cluster

        window : `float`
            window to use when clustering data points, will raise
            ValueError if `window > 0` is not satisfied

        Returns
        -------
        table : `EventTable`
            a new table that has had the clustering algorithm applied via
            slicing of the original

        Examples
        --------
        To cluster an `EventTable` (``table``) whose `index` is
        `end_time`, `window` is `0.1`, and maximize over `snr`:

        >>> table.cluster('end_time', 'snr', 0.1)
        """
        if window <= 0.0:
            msg = "Window must be a positive value"
            raise ValueError(msg)

        # If no rows, no need to cluster
        if len(self) == 0:
            return self.copy()

        # Generate index and rank vectors that are ordered
        orderidx = numpy.argsort(self[index])
        col = self[index][orderidx]
        param = self[rank][orderidx]

        # Find all points where the index vector changes by less than window
        clusterpoints = numpy.where(numpy.diff(col) <= window)[0]

        # If no such cluster points, no need to cluster
        if len(clusterpoints) == 0:
            return self.copy()

        # Divide points into clusters of adjacent points
        sublists = numpy.split(clusterpoints,
                               numpy.where(numpy.diff(clusterpoints) > 1)[0]+1)

        # Add end-points to each cluster and find the index of the maximum
        # point in each list
        padded_sublists = [numpy.append(s, numpy.array([s[-1]+1]))
                           for s in sublists]
        maxidx = [s[numpy.argmax(param[s])] for s in padded_sublists]

        # Construct a mask that removes all points within clusters and
        # replaces them with the maximum point from each cluster
        mask = numpy.ones_like(col, dtype=bool)
        mask[numpy.concatenate(padded_sublists)] = False
        mask[maxidx] = True

        return self[orderidx[mask]]
